__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

# flake8: noqa: F401

from os import environ
from typing import List

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from mpl_toolkits.mplot3d import Axes3D  # pylint: disable=unused-import
import pandas as pd


class ScatterPlot:

    MARKERS = ['x', 'o', '^', 'v', '.', '<', '1', '2', '3', '4']

    def __init__(self, marker_column: str, xyz_columns: List[str], image_file: str) -> None:
        self._marker_column = marker_column
        self._xyz_columns = xyz_columns
        self._image_file = image_file
        self._set_matplotlib_backend()
        self._axis = self._setup_3d_plot()

    def plot_data_point(self, row: pd.Series) -> None:
        marker = self.MARKERS[int(row[self._marker_column]) % 10]
        x_s, y_s, z_s = row[self._xyz_columns]
        self._axis.scatter(
            x_s, y_s, z_s, marker=marker,
            c=list(mcolors.TABLEAU_COLORS.keys())[int(row[self._marker_column]) % 10]
        )

    def show_or_save(self) -> None:
        if self._image_file is None:
            plt.show()
        else:
            plt.savefig(self._image_file, bbox_inches='tight')

    def _set_matplotlib_backend(self) -> None:
        if environ.get('DISPLAY'):
            matplotlib.use('qt5agg')
        else:
            if not self._image_file:
                raise ValueError('image_file must be set when running headless')
            matplotlib.use('Agg')

    @staticmethod
    def _setup_3d_plot() -> plt.axis:
        fig = plt.figure()
        axis = fig.add_subplot(111, projection='3d')
        axis.set_xlabel('0')
        axis.set_ylabel('1')
        axis.set_zlabel('2')
        return axis



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import pandas as pd


class TrackDataDuck:
    """
    Simulate TrackData objects to give mypy type information in TrackDataStorage
    without creating circular imports.
    """

    @property
    def data_frame(self) -> pd.DataFrame:
        """Quack, quack."""
        raise NotImplementedError()

    @property
    def hash(self) -> str:
        raise NotImplementedError()



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import shutil
from pathlib import Path
from time import time
from typing import Callable, Optional, Set

import pandas as pd

from .track_data_duck import TrackDataDuck

DEFAULT_STORE_EVERY = 100


class TrackDataStorage:

    def __init__(
            self, file_name: Path, quiet: bool = False, store_every: int = DEFAULT_STORE_EVERY
    ):
        self.file_name = file_name
        self.quiet = quiet
        self.tracks = self.load()
        self._track_buffer: Set[TrackDataDuck] = set()
        self.store_every = store_every

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._store_buffer()

    @property
    def num_buffered(self) -> int:
        return len(self._track_buffer)

    def load(self) -> pd.DataFrame:
        try:
            return pd.read_csv(self.file_name)
        except (FileNotFoundError, pd.errors.EmptyDataError):
            return pd.DataFrame({'hash': []})

    def add_track(self, track: TrackDataDuck) -> None:
        self._track_buffer.add(track)
        if len(self._track_buffer) >= self.store_every:
            self._store_buffer()

    def add_dataframe(self, tracks: pd.DataFrame) -> None:
        hashes_to_add = set(tracks['hash']) - set(self.tracks['hash'])
        tracks_to_add = tracks.loc[tracks['hash'].isin(hashes_to_add)]
        self.tracks = self.tracks.append(tracks_to_add)

    def _store_buffer(self):
        self.store_set(self._track_buffer)
        self._track_buffer = set()

    def store_set(self, tracks: Set[TrackDataDuck]) -> pd.DataFrame:
        if tracks:
            return self.store_dataframe(pd.concat(track.data_frame for track in tracks))
        return self.tracks

    def store_dataframe(self, tracks: pd.DataFrame) -> pd.DataFrame:
        if not tracks.empty:
            self.timed_if_wanted(self.actual_store, tracks)
        return self.tracks

    def actual_store(self, tracks):
        self.add_dataframe(tracks)
        if self.file_name.exists():
            shutil.copy(self.file_name, self.file_name.with_suffix('.old'))
        self.tracks.to_csv(str(self.file_name), compression='infer')

    def from_hash(self, file_hash: str) -> Optional[pd.DataFrame]:
        try:
            found = self.tracks.loc[self.tracks['hash'] == file_hash]
            if found.shape[0] == 0:
                candidates = [track for track in self._track_buffer if track.hash == file_hash]
                if candidates:
                    return candidates[0].data_frame
                return None
            return found
        except KeyError:
            return None

    def timed_if_wanted(self, func: Callable, *args):
        start_time = 0.
        if not self.quiet:
            start_time = time()
            print(f"    Writing tracks to {self.file_name}...", end=' ', flush=True)
        func(*args)
        if not self.quiet:
            elapsed = time() - start_time
            print(f"done ({self.tracks.shape[0]} tracks, {elapsed:.1f}s)")



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from typing import Tuple

import numpy as np
import pandas as pd
from sklearn.decomposition import PCA

from mafe.scan import TrackData


def pca(data: pd.DataFrame, total_variance_ratio: float = 0.99) -> Tuple[PCA, np.ndarray]:
    reducer = PCA(total_variance_ratio, svd_solver='full')
    transformed = reducer.fit_transform(data.to_numpy())
    return reducer, transformed


class Reducer:
    def __init__(
            self, tracks_csv: str, total_variance_ratio: float = 0.99,
            num_cols: Tuple[str, ...] = None, text_cols: Tuple[str, ...] = None
    ) -> None:
        self.tracks = pd.read_csv(tracks_csv)
        self.total_variance_ratio = total_variance_ratio
        self.num_cols = TrackData.numerical_columns() if num_cols is None else num_cols
        self.text_cols = TrackData.STRING_COLUMNS if text_cols is None else text_cols
        self.reducer: PCA = None

    def run(self) -> pd.DataFrame:
        numbers_only = self.tracks[list(self.num_cols)].dropna(axis='columns')
        text = self.tracks[list(self.text_cols)]
        self.reducer, result = pca(numbers_only, self.total_variance_ratio)
        reduced = pd.DataFrame(result)
        reduced[list(self.text_cols)] = text
        return reduced

    @property
    def explained_variance_ratio(self):
        return self.reducer.explained_variance_ratio_



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from .calculate_distances import DistanceCalculator, distances  # noqa: F401
from .normalize import normalize, norm_zero_max, norm_min_max, norm_mean  # noqa: F401
from .pca import pca  # noqa: F401
from .cluster import Cluster  # noqa: F401



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from typing import Tuple

import audio_metadata

from .types import Metadata

METADATA_COLUMNS = ('artist', 'title', 'album')


def read_track_metadata(file_name: str) -> Metadata:
    try:
        metadata = audio_metadata.load(file_name)
        return dict(_extract_tag(metadata, tag) for tag in METADATA_COLUMNS)
    except audio_metadata.AudioMetadataException as problem:
        raise RuntimeError from problem


def _extract_tag(metadata: audio_metadata.Format, tag: str) -> Tuple[str, str]:
    return tag, metadata['tags'].get(tag, ['...'])[0]



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

# suppress warnings since warnings import interferes with MP3 loading
from contextlib import redirect_stderr
from functools import lru_cache
from hashlib import sha3_224
from logging import info
from pathlib import Path
from typing import Callable, Optional, Tuple

from scipy import stats
import numpy as np
import pandas as pd
import librosa

from .track_metadata import METADATA_COLUMNS, read_track_metadata
from .track_data_duck import TrackDataDuck
from .track_data_storage import TrackDataStorage
from .types import Metadata


class TrackTooLong(ValueError):
    def __init__(self, length: float):
        super().__init__(f"{length:.0f}s > {TrackData.max_track_duration}s")


class TrackData(TrackDataDuck):

    FEATURE_SIZES = dict(
        chroma_stft=12, chroma_cqt=12, chroma_cens=12, tonnetz=6, mfcc=20, rms=1, zcr=1,
        spectral_centroid=1, spectral_bandwidth=1, spectral_contrast=7, spectral_rolloff=1
    )
    MOMENTS = ('mean', 'std', 'skew', 'kurtosis', 'median', 'min', 'max')
    STRING_COLUMNS = ('hash', 'file_name') + METADATA_COLUMNS
    META_COLUMNS = STRING_COLUMNS + ('tempo', 'duration')

    max_track_duration = 7200

    @classmethod
    def set_max_track_duration(cls, max_duration: int):
        cls.max_track_duration = max_duration

    def __init__(self, storage: TrackDataStorage, track: Path):
        self.track = track
        self.metadata: Metadata = None
        self.feature_series: pd.Series = None
        self._data_frame: pd.DataFrame = None
        with redirect_stderr(None):
            self.load_track(storage, track)

    @property
    def hash(self) -> Optional[str]:
        return str(self.metadata.get('hash')) if self.metadata is not None else None

    @property
    def data_frame(self) -> pd.DataFrame:
        return self._data_frame

    def load_track(self, storage: TrackDataStorage, track: Path):
        metadata = self.read_metadata()
        data_frame = storage.from_hash(str(metadata['hash']))
        if data_frame is not None:
            self._data_frame = data_frame
        else:
            self.load_audio(metadata, track)

    def load_audio(self, metadata, track):
        waveform, sample_rate = librosa.load(str(track))
        metadata.update(self.expensive_metadata(waveform, sample_rate))
        self.metadata = metadata
        self.feature_series = pd.Series(index=TrackData.columns(), dtype=np.float32)
        self.compute_features(waveform, sample_rate)
        self._data_frame = self.feature_series.to_frame().transpose().assign(
            **self.metadata
        )

    @staticmethod
    @lru_cache()
    def columns() -> pd.Index:
        return pd.Index(TrackData.META_COLUMNS + TrackData.stats_columns())

    @staticmethod
    @lru_cache()
    def numerical_columns():
        return ('tempo', 'duration') + TrackData.stats_columns()

    @staticmethod
    @lru_cache()
    def stats_columns() -> Tuple[str, ...]:
        return tuple(sorted([
            f"{name}_{moment}_{i + 1:02d}"
            for name, size in TrackData.FEATURE_SIZES.items()
            for moment in TrackData.MOMENTS
            for i in range(size)
        ]))

    def compute_features(self, waveform: np.ndarray, sample_rate: float) -> None:
        self.feature_stats(
            'zcr', librosa.feature.zero_crossing_rate(waveform, frame_length=2048, hop_length=512)
        )

        cqt = np.abs(librosa.cqt(
            waveform, sr=sample_rate, hop_length=512,
            bins_per_octave=12, n_bins=7 * 12, tuning=None
        ))
        self.feature_stats(
            'chroma_cqt', librosa.feature.chroma_cqt(C=cqt, n_chroma=12, n_octaves=7)
        )
        chroma_cens = librosa.feature.chroma_cens(C=cqt, n_chroma=12, n_octaves=7)
        self.feature_stats('chroma_cens', chroma_cens)
        self.feature_stats('tonnetz', librosa.feature.tonnetz(chroma=chroma_cens))

        stft = np.abs(librosa.stft(waveform, n_fft=2048, hop_length=512))
        self.feature_stats('chroma_stft', librosa.feature.chroma_stft(S=stft ** 2, n_chroma=12))
        self.feature_stats('rms', librosa.feature.rms(S=stft))
        self.feature_stats('spectral_centroid', librosa.feature.spectral_centroid(S=stft))
        self.feature_stats('spectral_bandwidth', librosa.feature.spectral_bandwidth(S=stft))
        self.feature_stats(
            'spectral_contrast', librosa.feature.spectral_contrast(S=stft, n_bands=6)
        )
        self.feature_stats('spectral_rolloff', librosa.feature.spectral_rolloff(S=stft))

        mel = librosa.feature.melspectrogram(sr=sample_rate, S=stft ** 2)
        self.feature_stats('mfcc', librosa.feature.mfcc(S=librosa.power_to_db(mel), n_mfcc=20))

    def feature_stats(self, name: str, values: np.ndarray) -> None:
        self.set_feature(name, values, "mean", np.mean)
        self.set_feature(name, values, "std", np.std)
        self.set_feature(name, values, "skew", stats.skew)
        self.set_feature(name, values, "kurtosis", stats.kurtosis)
        self.set_feature(name, values, "median", np.median)
        self.set_feature(name, values, "min", np.min)
        self.set_feature(name, values, "max", np.max)

    def set_feature(self, name: str, values: np.ndarray, moment: str, function: Callable) -> None:
        for i, value in enumerate(function(values, axis=1)):
            self.feature_series[f"{name}_{moment}_{i + 1:02d}"] = value

    def read_metadata(self) -> Metadata:
        with self.track.open('rb') as file:
            file_hash = sha3_224(file.read()).hexdigest()[:32]
        duration = librosa.get_duration(filename=str(self.track))
        if duration > self.max_track_duration:
            raise TrackTooLong(duration)
        return dict(file_name=self.track.name, hash=file_hash, duration=duration)

    def expensive_metadata(self, waveform: np.ndarray, sample_rate: float) -> Metadata:
        tempo, _ = librosa.beat.beat_track(y=waveform, sr=sample_rate)
        metadata = dict(tempo=tempo)
        try:
            with redirect_stderr(None):
                metadata.update(read_track_metadata(str(self.track)))
        except (RuntimeError, ValueError) as problem:
            info(problem)
        return metadata
        # last.fm stats
        # - loved
        # - num plays
        # - length of period of plays



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from pathlib import Path
from shutil import get_terminal_size
from typing import List, Optional

from zlib import error as zlib_error
from audioread.exceptions import NoBackendError
import magic
import pandas as pd

from .track_data import TrackData, TrackTooLong
from .track_data_storage import DEFAULT_STORE_EVERY, TrackDataStorage

AUDIO_EXTENSIONS = ('mp3', 'ogg', 'flac', 'wav', 'm4a', 'aif', 'aiff')
RED = '\033[91m'
NORMAL = '\033[0m'


class Scanner:

    def __init__(
            self, file_name: str, base_folders: List[str], quiet: bool,
            store_every: int = DEFAULT_STORE_EVERY
    ) -> None:
        self.csv_file = file_name
        self.files = [
            file for folder in base_folders
            for file in Path(folder).rglob('*')
            if file.is_file()
            if is_audio(file)
        ]
        self.quiet = quiet
        self.store_every = store_every
        self.done = 0
        self.failed = 0
        self.interrupted = False

    def run(self) -> pd.DataFrame:
        with TrackDataStorage(Path(self.csv_file), self.quiet, self.store_every) as storage:
            try:
                for track in sorted(self.files):
                    self.process_track(storage, track)
            except KeyboardInterrupt:
                pass
        return storage.tracks

    def process_track(self, storage: TrackDataStorage, track: Path) -> None:
        try:
            features = TrackData(storage, track)
        except (
                FileNotFoundError, IndexError, NoBackendError, TrackTooLong,
                UnicodeEncodeError, UserWarning, ValueError, ZeroDivisionError, zlib_error
        ) as problem:
            self.log_problem(track, problem)
            return
        except KeyboardInterrupt:
            if self.interrupted:
                raise
            print(
                f"{RED}Interrupting {track}.{NORMAL} "
                f"A second keyboard interrupt will terminate the scan.")
            self.interrupted = True
            features = None

        self.done += 1
        self.store_track(storage, features)

    def log_problem(self, track: Path, problem: Exception) -> None:
        self.failed += 1
        track_part = track.name.encode('utf-8', 'surrogateescape').decode('utf-8', 'replace')
        problem_part = f"{type(problem).__name__}: {problem}"
        progress = self.progress()
        length = len(problem_part) + len(progress) + 7
        self.print(
            f"{cut(track_part, term_width() - length)} "
            f"{RED}{problem_part}{NORMAL}   [{progress}]"
        )

    def store_track(self, storage: TrackDataStorage, track: Optional[TrackData]) -> None:
        if track is None:
            return
        self.print_info(track)
        storage.add_track(track)

    def print_info(self, track: TrackData) -> None:
        hash_part = track.data_frame.iloc[0]['hash'][:8]
        duration = f"{int(track.data_frame.iloc[0]['duration']):4}s"
        tempo = f"{int(track.data_frame.iloc[0]['tempo']):3}BPM"
        progress = self.progress()
        length = len(hash_part) + len(duration) + len(tempo) + len(progress) + 9
        self.print(
            f"{cut(track.data_frame.iloc[0]['file_name'], term_width() - length)} "
            f"{hash_part} {duration}@{tempo}   [{progress}]"
        )

    def progress(self) -> str:
        return f"{self.done:3} ok/{self.failed:2} fail/ {len(self.files)} tot"

    def print(self, *args, **kwargs):
        if not self.quiet:
            print(*args, **kwargs)


def is_audio(track: Path):
    if track.suffix[1:].lower() in AUDIO_EXTENSIONS:
        return True
    file_magic = magic.from_file(str(track), mime=True)
    return file_magic.startswith('audio/')


def cut(string: str, length: int, min_width: int = 10) -> str:
    if len(string) <= length:
        return string.ljust(length)
    length = max(length, min_width)
    return string[:length - 3] + '...'


def term_width() -> int:
    return get_terminal_size((80, 20))[0]



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import numpy as np
import pandas as pd

from mafe.scan import TrackData
from .normalize import normalize


def distances(tracks: pd.DataFrame, verbose: bool = False) -> pd.DataFrame:
    this_hash, other_hash, distance = [], [], []
    for this_index, this_row in enumerate(tracks[TrackData.columns()].values):
        for other_index, other_row in enumerate(tracks[TrackData.columns()].values):
            if other_index <= this_index:
                continue
            difference = np.subtract(this_row[5:], other_row[5:])
            this_hash.append(this_row[0])
            other_hash.append(other_row[0])
            distance.append(np.linalg.norm(difference))
        if verbose:
            print(
                f"{this_index + 1}/{tracks.shape[0]} len={len(distance)}",
                end="\r", flush=True
            )
    if verbose:
        print()
    return pd.DataFrame.from_dict(
        {'hash_1': this_hash, 'hash_2': other_hash, 'process': distance}
    )


class DistanceCalculator:  # pylint: disable=too-few-public-methods

    def __init__(self, tracks_csv: str, output_filename: str) -> None:
        self.tracks = pd.read_csv(tracks_csv)
        self.output_filename = output_filename

    def run(self, verbose: bool):
        normalized = normalize(self.tracks)
        distances(normalized, verbose=verbose).to_csv(self.output_filename)



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from typing import Callable, Tuple

import pandas as pd

from mafe.scan import TrackData

NormalizationFunction = Callable[[pd.DataFrame], pd.DataFrame]


def norm_min_max(frame: pd.DataFrame) -> pd.DataFrame:
    return (frame - frame.min()) / (frame.max() - frame.min())


def norm_zero_max(frame: pd.DataFrame) -> pd.DataFrame:
    zero_min = frame.min().apply(lambda x: min(x, 0))
    return (frame - zero_min) / (frame.max() - zero_min)


def norm_mean(frame: pd.DataFrame) -> pd.DataFrame:
    return (frame - frame.mean()) / frame.std()


NORMALIZATION_FUNCTIONS = {
    'mean': norm_mean,
    'min_max': norm_min_max,
    'zero_max': norm_zero_max
}


def normalize(
        tracks: pd.DataFrame, method: str = 'mean', text_cols: Tuple[str, ...] = None
) -> pd.DataFrame:
    func = NORMALIZATION_FUNCTIONS[method]
    if text_cols is None:
        text_cols = TrackData.STRING_COLUMNS
    numbers_only = tracks.drop(list(text_cols), axis='columns') if text_cols else tracks
    text = tracks[list(text_cols)]
    normalized = func(numbers_only)
    normalized[list(text_cols)] = text
    return normalized



#! /usr/bin/env python3
__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from typing import List

import click
import pandas as pd

from mafe.process import DistanceCalculator, normalize as proc_normalize, Cluster
from mafe.process.pca import Reducer
from mafe.scan import Scanner, TrackData, DEFAULT_STORE_EVERY


@click.group()
@click.option(
    '--tracks-csv', '-t', required=True, help='CSV file containing tracks'
)
@click.option(
    '--output', '-o', help='CSV file containing processed track data'
)
@click.pass_context
def cli(ctx, tracks_csv: str, output: str):
    ctx.ensure_object(dict)
    ctx.obj['tracks_csv'] = tracks_csv
    ctx.obj['output'] = output


@cli.command()
@click.pass_context
@click.option(
    '--base-folders', '-f', multiple=True, required=True, help='Directory to scan'
)
@click.option(
    '--max-track-length', '-m', type=int,
    help='Maximum track length, in seconds (longer tracks are ignored)'
)
@click.option(
    '--quiet', '-q', is_flag=True, default=False, help='Suppress warnings and progress messages'
)
@click.option(
    '--store-every', '-s', type=int, default=DEFAULT_STORE_EVERY, help='Store every n tracks'
)
def scan(
        ctx: click.Context, base_folders: List[str], max_track_length: int,
        quiet: bool, store_every: int
) -> None:
    if max_track_length is not None:
        TrackData.set_max_track_duration(max_track_length)

    scanner = Scanner(ctx.obj['tracks_csv'], base_folders, quiet, store_every)
    scanner.run()


@cli.command()
@click.pass_context
@click.option(
    '--method', '-m', type=click.Choice(['mean', 'min_max', 'zero_max'], case_sensitive=False),
    default='mean', help='Normalization method'
)
def normalize(ctx: click.Context, method: str) -> None:
    data = pd.read_csv(ctx.obj['tracks_csv'])
    normalized = proc_normalize(data, method=method)
    normalized.to_csv(ctx.obj['output'])


@cli.command()
@click.pass_context
def pca(ctx: click.Context) -> None:
    reducer = Reducer(ctx.obj['tracks_csv'])
    reducer.run().to_csv(ctx.obj['output'])


@cli.command()
@click.pass_context
@click.option(
    '--verbose', '-v', is_flag=True, default=False, help='Show progress info'
)
def distance(ctx: click.Context, verbose: bool) -> None:
    calculator = DistanceCalculator(ctx.obj['tracks_csv'], ctx.obj['output'])
    calculator.run(verbose=verbose)


@cli.command()
@click.option(
    '--num-clusters', '-n', required=True, type=int, help='Number of clusters to compute'
)
@click.option(
    '--max-iter', '-m', type=int, default=300, help='Maximum iterations to run'
)
@click.option(
    '--visualize', '-V', is_flag=True, default=False, help='Visualize clusters'
)
@click.option(
    '--num-samples', '-N', type=int, default=None, help='Number of points for cluster visualization'
)
@click.option(
    '--image-file', '-I', default=None, help='Cluster visualization image file'
)
@click.pass_context
def cluster(  # pylint: disable=too-many-arguments
        ctx: click.Context,
        num_clusters: int, max_iter: int, visualize: bool, num_samples: int, image_file: str
) -> None:
    clusterer = Cluster(ctx.obj['tracks_csv'], num_clusters)
    clusterer.run(max_iter).to_csv(ctx.obj['output'])
    if visualize:
        clusterer.visualize(n_samples=num_samples, image_file=image_file)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mafe', 'mafe.process', 'mafe.scan']

package_data = \
{'': ['*']}

install_requires = \
['PyQt5>=5.15.0,<6.0.0',
 'audio-metadata>=0.11.1,<0.12.0',
 'audioread>=2.1.8,<3.0.0',
 'click>=7.1.2,<8.0.0',
 'librosa>=0.8.0,<0.9.0',
 'matplotlib>=3.2.1,<4.0.0',
 'numba>=0.50,<0.51',
 'numpy>=1.18.5,<2.0.0',
 'pandas>=1.0.4,<2.0.0',
 'python-magic>=0.4.18,<0.5.0',
 'scikit-learn>=0.23.1,<0.24.0',
 'scipy>=1.4.1,<2.0.0']

entry_points = \
{'console_scripts': ['mafe = mafe.mafe:cli']}

setup_kwargs = {
    'name': 'mafe',
    'version': '0.3.0',
    'description': 'Music Audio Feature Extractor',
    'long_description': '# Music Audio Feature Extractor\n\n## Installation\n```shell script\n$ pip install mafe\n```\n\n## Typical usages\n```shell script\n# scan music tracks to extract raw set of features\n$ mafe -t scanned.csv.bz2 scan -f [MUSIC_DIR]\n# run normalization on extracted features\n$ mafe -t scanned.csv.bz2 -o normalized.csv.bz2 normalize\n# create table of distances between tracks\n$ mafe -t normalized.csv.bz2 -o distances.csv.bz2 distance\n# find clusters of similar tracks\n$ mafe -t normalized.csv.bz2 -o clustered.csv.bz2 cluster -n 4\n# run dimensionality reduction, keeping only the most distinctive features\n$ mafe -t normalized.csv.bz2 -o reduced.csv.bz2 pca\n# run clustering on distinct features, creating a visualization of the clusters\n$ mafe -t reduced.csv.bz2 -o clustered_reduced.csv.bz2 cluster -n 4 -V -I cluster.png\n```\n\n## Command line options\n```shell script\n$ mafe --help\nUsage: mafe [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  -t, --tracks-csv TEXT  CSV file containing tracks  [required]\n  -o, --output TEXT      CSV file containing distances between the tracks\n                         [required]\n  --help                 Show this message and exit.\n\nCommands:\n  cluster\n  distance\n  normalize\n  pca\n  scan\n```\n',
    'author': 'Lene Preuss',
    'author_email': 'lene.preuss@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/lilacashes/mafe',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from typing import Tuple

import pandas as pd
from sklearn.cluster import KMeans

from mafe.process.scatter_plot import ScatterPlot
from mafe.scan import TrackData


def k_means(tracks: pd.DataFrame, n_clusters: int, max_iter: int) -> KMeans:
    return KMeans(n_clusters=n_clusters, max_iter=max_iter).fit(tracks)


class Cluster:

    def __init__(
            self, tracks_csv: str, n_clusters: int, text_cols: Tuple[str, ...] = None
    ) -> None:
        self.tracks = pd.read_csv(tracks_csv)
        self.text_cols = list(TrackData.STRING_COLUMNS) if text_cols is None else text_cols
        self.n_clusters = n_clusters

    def run(self, n_iter: int = 300) -> pd.DataFrame:
        numbers_only = self.tracks.drop(self.text_cols, axis='columns').dropna(axis='columns')
        kmeans = k_means(numbers_only, self.n_clusters, n_iter)
        self.tracks.insert(1, 'cluster', kmeans.labels_, True)
        return self.tracks

    def visualize(self, n_samples: int = None, image_file: str = None) -> None:
        visualizer = ScatterPlot('cluster', ['0', '1', '2'], image_file)
        data = self._reduce_data(n_samples)

        for _, row in data.iterrows():
            visualizer.plot_data_point(row)

        visualizer.show_or_save()

    def _reduce_data(self, n_samples: int) -> pd.DataFrame:
        if 'cluster' not in self.tracks.columns:
            raise ValueError('Run clustering before visualizing')
        try:
            data = self.tracks[['cluster', '0', '1', '2']]
        except KeyError as error:
            raise ValueError('Run dimensionality reduction before visualizing') from error

        if n_samples is not None and n_samples < len(data):
            data = data.sample(n=n_samples)
        return data
